//
// Copyright (c) 2022, MindFusion LLC - Bulgaria.
//

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using MindFusion.Diagramming;
using MindFusion.Drawing;
using MindFusion.Layout;


namespace MindFusion.Diagramming.WinForms.Samples.CS.MoleculeViewer
{
	public partial class MainForm : Form
	{
		public MainForm()
		{
			InitializeComponent();

			diagram.DefaultShape = Shapes.Ellipse;

			twoLinks = new MindFusion.Drawing.Pen(Color.FromArgb(45, 57, 86), 1f);
			twoLinks.CompoundArray = new float[] { 0, 0.3f, 0.7f, 1 };

            brush1 = new LinearGradientBrush(Color.FromArgb(224, 233, 233), Color.FromArgb(90, 121, 165), 44);
            brush2 = new LinearGradientBrush(Color.FromArgb(224, 233, 233), Color.FromArgb(156, 170, 198), 44);

			linkers = new ArrayList();
			queue = new Queue<Linker>();
			linkedNodes = new Queue<Linker>();

			diagram.DrawNode += diagram_DrawNode;
		}

		void diagram_DrawNode(object sender, DrawNodeEventArgs e)
		{
			System.Drawing.Drawing2D.GraphicsPath x =
				new System.Drawing.Drawing2D.GraphicsPath();

			x.AddEllipse(e.Bounds);

            Color surrounding = (bool)e.Node.Tag ? Color.FromArgb(90, 121, 165) : Color.FromArgb(156, 170, 198);

			System.Drawing.Drawing2D.PathGradientBrush p =
				new System.Drawing.Drawing2D.PathGradientBrush(x);
            p.CenterColor = Color.FromArgb(224, 233, 233);
			p.SurroundColors = new Color[] { surrounding };

			e.Graphics.FillEllipse(p, e.Bounds);

			p.Dispose();
			x.Dispose();

			System.Drawing.Pen pp = new System.Drawing.Pen(surrounding, 0);

			e.Graphics.DrawEllipse(pp, e.Bounds);

			pp.Dispose();
		}

		private void Form1_Load(object sender, EventArgs e)
		{
			diagramView3D.Focus();
		}

		private void OnLoadEthaneClick(object sender, EventArgs e)
		{
			diagram.ClearAll();
			LoadEthane();
			diagramView3D.ResetProjection();
			diagramView3D.Focus();
		}

		private void OnLoadBenzeneClick(object sender, EventArgs e)
		{
			diagram.ClearAll();
			LoadBenzene();
			diagramView3D.ResetProjection();
			diagramView3D.Focus();
		}

		private void OnLoadBuckminsterfullereneClick(object sender, EventArgs e)
		{
			diagram.ClearAll();
			LoadBuckminsterfullerene();
			diagramView3D.ResetProjection();
			diagramView3D.Focus();
		}

		private void LoadEthane()
		{
			float[,] xyz = 
				{
					{0, 0, 0},					
					{-18, 29.8f, 2.93f},
					{-18, -12.33f, -27.34f},					
					{-18, -17.51f, 24.35f},
					{30, 0, 0},
					{38, -29.8f, -2.93f},
					{38, 12.33f, 27.34f},					
					{38, 17.511f, -24.35f}
				};

			ShapeNode zNode = null;
			ShapeNode cNode = null;

			for (int i = 0; i < xyz.Length / 3; i++)
			{
				if (i % 4 == 0)
				{
					if (i == 0)
					{
						cNode = diagram.Factory.CreateShapeNode(0, 0, 20, 20);
						cNode.CustomDraw = CustomDraw.Full;
						zNode = cNode;
					}
					else
					{
						zNode = diagram.Factory.CreateShapeNode(0, 0, 20, 20);
						zNode.CustomDraw = CustomDraw.Full;
						DiagramLink link = diagram.Factory.CreateDiagramLink(cNode, zNode);
						link.Pen = twoLinks;
						cNode = zNode;
					}
				}
				else
				{
					zNode = diagram.Factory.CreateShapeNode(0, 0, 15, 15);
					zNode.CustomDraw = CustomDraw.Full;
					diagram.Factory.CreateDiagramLink(cNode, zNode);
				}

				zNode.LayoutTraits[View3D.Position] = new Point3D(xyz[i, 0], xyz[i, 1], xyz[i, 2]);
				zNode.Text = i % 4 == 0 ? "C" : "H" + i;
				zNode.Brush = i % 4 == 0 ? brush2 : brush1;
				zNode.Tag = i % 4 == 0;
			}
		}

		private void LoadBenzene()
		{
			ShapeNode zNode;

			float[,] xyz =
			{
				{3, 33},
				{5, 63},
				{30, 13},
				{53, 32},
				{27, -19},
				{48, -41},
				
				{-3, -33},
				{-5, -63},
				{-30, -13},
				{-53, -32},
				{-27, 19},
				{-48, 41}
			};

			DiagramNode buffNodeC = null;
			DiagramNode firstNode = null;



			bool isTwoLinks = false;
			for (int i = 0; i < 12; i++)
			{
				zNode = diagram.Factory.CreateShapeNode(0, 0, 23, 23);
				zNode.CustomDraw = CustomDraw.Full;
				if (i % 2 == 0)
				{
					zNode.Text = "C";
					zNode.Brush = brush2;
					zNode.Tag = true;
					if (buffNodeC != null)
					{
						if (isTwoLinks)
						{
							DiagramLink link = diagram.Factory.CreateDiagramLink(buffNodeC, zNode);
						}
						else
						{
							DiagramLink link = diagram.Factory.CreateDiagramLink(buffNodeC, zNode);
							link.Pen = twoLinks;
						}

						buffNodeC = zNode;
						isTwoLinks = !isTwoLinks;
					}
					else
					{
						buffNodeC = zNode;
						firstNode = zNode;
					}
				}
				else
				{
					zNode.Text = "H";
					zNode.Brush = brush1;
					zNode.Tag = false;
					diagram.Factory.CreateDiagramLink(buffNodeC, zNode);
				}

				zNode.LayoutTraits[View3D.Position] = new Point3D(xyz[i, 0], xyz[i, 1], 0);
			}

			diagram.Factory.CreateDiagramLink(buffNodeC, firstNode);
		}

		private void LoadBuckminsterfullerene()
		{
			float[,] xyz =
			{
				{36.795f, 0.00000000f, 99.435f},
				{11.37f, 34.992f, 99.435f},
				{-29.766f, 21.627f, 99.435f},
				{-29.766f, -21.627f, 99.435f},
				{11.37f, -34.992f, 99.435f},
				{102.252f, 21.627f, 17.844f},
				{83.853f, 34.992f, 54.639f},
				{72.483f, 0.00000000f, 77.379f},
				{83.853f, -34.992f, 54.639f},
				{102.252f, -21.627f, 17.844f},
				{11.028f, 103.929f, 17.844f},
				{-7.368f, 90.564f, 54.639f},
				{22.398f, 68.937f, 77.379f},
				{59.193f, 68.937f, 54.639f},
				{52.167f, 90.564f, 17.844f},
				{-95.436f, 42.606f, 17.844f},
				{-88.407f, 20.979f, 54.639f},
				{-58.641f, 42.606f, 77.379f},
				{-47.271f, 77.598f, 54.639f},
				{-70.011f, 77.598f, 17.844f},
				{-70.011f, -77.598f, 17.844f},
				{-47.271f, -77.598f, 54.639f},
				{-58.641f, -42.606f, 77.379f},
				{-88.407f, -20.979f, 54.639f},
				{-95.436f, -42.606f, 17.844f},
				{52.167f, -90.564f, 17.844f},
				{59.193f, -68.937f, 54.639f},
				{22.398f, -68.937f, 77.379f},
				{-7.368f, -90.564f, 54.639f},
				{11.028f, -103.929f, 17.844f},
				{29.766f, 21.627f, -99.435f},
				{-11.37f, 34.992f, -99.435f},
				{-36.795f, 0.00000000f, -99.435f},
				{-11.37f, -34.992f, -99.435f},
				{29.766f, -21.627f, -99.435f},
				{70.011f, 77.598f, -17.844f},
				{47.271f, 77.598f, -54.639f},
				{58.641f, 42.606f, -77.379f},
				{88.407f, 20.979f, -54.639f},
				{95.436f, 42.606f, -17.844f},
				{-52.167f, 90.564f, -17.844f},
				{-59.193f, 68.937f, -54.639f},
				{-22.398f, 68.937f, -77.379f},
				{7.368f, 90.564f, -54.639f},
				{-11.028f, 103.929f, -17.844f},
				{-102.252f, -21.627f, -17.844f},
				{-83.853f, -34.992f, -54.639f},
				{-72.483f, 0.00000000f, -77.379f},
				{-83.853f, 34.992f, -54.639f},
				{-102.252f, 21.627f, -17.844f},
				{-11.028f, -103.929f, -17.844f},
				{7.368f, -90.564f, -54.639f},
				{-22.398f, -68.937f, -77.379f},
				{-59.193f, -68.937f, -54.639f},
				{-52.167f, -90.564f, -17.844f},
				{95.436f, -42.606f, -17.844f},
				{88.407f, -20.979f, -54.639f},
				{58.641f, -42.606f, -77.379f},
				{47.271f, -77.598f, -54.639f},
				{70.011f, -77.598f, -17.844f}
			};

			// create the nodes
			ShapeNode zNode = null;
			for (int i = 0; i < xyz.Length / 3; i++)
			{
				zNode = diagram.Factory.CreateShapeNode(0, 0, 15, 15);
				zNode.CustomDraw = CustomDraw.Full;
				zNode.LayoutTraits[View3D.Position] = new Point3D(xyz[i, 0], xyz[i, 1], xyz[i, 2]);
				zNode.Brush = brush2;
				zNode.Tag = true;
				zNode.Text = "C";

				Linker linker = new Linker(zNode);
				linkers.Add(linker);
				if (queue.Count == 0)
					queue.Enqueue(linker);
			}

			SetLinks();
			queue.Clear();
			linkedNodes.Clear();
			linkers.Clear();
		}

		private void SetLinks()
		{
			while (queue.Count != 0)
			{
				Linker linker = queue.Dequeue();
				linkedNodes.Enqueue(linker);
				ShapeNode node = linker.CurrentNode;

				double[] dists = new double[4];
				Linker[] lks = new Linker[4];
				for (int j = 0; j < dists.Length; j++)
				{
					dists[j] = -1;
					lks[j] = null;
				}

				int n = 0;

				// find the smallest distances
				for (int ii = 0; ii < linkers.Count; ii++)
				{
					Linker lk = (Linker)linkers[ii];
					ShapeNode nd = lk.CurrentNode;
					if (node.Equals(nd))
						continue;

					// calculate the distance between the lks
					double d = CalcDistance((Point3D)node.LayoutTraits[View3D.Position], (Point3D)nd.LayoutTraits[View3D.Position]);

					// sorting by distance
					if (n >= dists.Length)
					{
						dists[dists.Length - 1] = d;
						lks[dists.Length - 1] = lk;
						Sort(dists, lks);
					}
					else
					{
						for (int m = 0; m < dists.Length; m++, n++)
							if (dists[m] == -1)
							{
								dists[m] = d;
								lks[m] = lk;
								Sort(dists, lks);
								n++;
								break;
							}
					}
				}

				// set the node's links
				for (int ii = 0; ii < dists.Length - 1; ii++)
				{
					if (lks[ii].Counter == 3 || linkedNodes.Contains(lks[ii]))
						continue;

					if (linker.DoubleLink == null && lks[ii].DoubleLink == null)
					{
						DiagramLink link = diagram.Factory.CreateDiagramLink(node, lks[ii].CurrentNode);
						link.Pen = twoLinks;
						linker.DoubleLink = lks[ii].CurrentNode;

						lks[ii].DoubleLink = node;
					}
					else
					{
						diagram.Factory.CreateDiagramLink(node, lks[ii].CurrentNode);
						linker.SingleLink(lks[ii].CurrentNode);
						lks[ii].SingleLink(linker.CurrentNode);
					}

					if (!queue.Contains(lks[ii]))
						queue.Enqueue(lks[ii]);
				}
			}
		}

		// Bubble sort
		private void Sort(double[] dist, Linker[] lks)
		{
			for (int i = 0; i < dist.Length - 1; i++)
			{
				for (int j = 0; j < dist.Length - 1 - i; j++)
				{
					if (dist[j + 1] < dist[j])
					{
						double tmpD = dist[j];
						dist[j] = dist[j + 1];
						dist[j + 1] = tmpD;

						Linker tmpL = lks[j];
						lks[j] = lks[j + 1];
						lks[j + 1] = tmpL;
					}
				}
			}
		}

		private double CalcDistance(Point3D src, Point3D dst)
		{
			double x = Math.Abs(src.X - dst.X);
			double y = Math.Abs(src.Y - dst.Y);
			double z = Math.Abs(src.Z - dst.Z);

			return Math.Sqrt(x * x + y * y + z * z);
		}

		MindFusion.Drawing.Pen twoLinks = null;
		LinearGradientBrush brush1;
		LinearGradientBrush brush2;

		ArrayList linkers;
		Queue<Linker> queue;
		Queue<Linker> linkedNodes;
	}
}